import os
import json
import pandas as pd
from report.entity.domain_config import DomainConfig
from report.elasticsearch_rule_converter import *

def gen_domain_config(data_folder,report_file):
    with open(data_folder + "/domain_config.txt") as f:
        config_txt = f.read()
        config = eval(config_txt)
        #print("ES version:" + str(config['ElasticsearchVersion']))
        version = config["ElasticsearchVersion"]["Options"]
        dn_intance_type = config["ElasticsearchClusterConfig"]["Options"]["InstanceType"]
        dn_instance_count = config["ElasticsearchClusterConfig"]["Options"]["InstanceCount"]
        dn_zone_awareness = config["ElasticsearchClusterConfig"]["Options"]["ZoneAwarenessEnabled"]
        dn_zone_number = 0
        if dn_zone_awareness:
            dn_zone_number = config["ElasticsearchClusterConfig"]["Options"]["ZoneAwarenessConfig"]["AvailabilityZoneCount"]
        dedicated_master = config["ElasticsearchClusterConfig"]["Options"]["DedicatedMasterEnabled"]
        dedicated_master_type = ""
        dedicated_master_count = 0
        if "DedicatedMasterType" in config["ElasticsearchClusterConfig"]["Options"]:
            dedicated_master_type = config["ElasticsearchClusterConfig"]["Options"]["DedicatedMasterType"]
            dedicated_master_count = config["ElasticsearchClusterConfig"]["Options"]["DedicatedMasterCount"]
        ebs = config["EBSOptions"]["Options"]["VolumeType"]
        ebs_volume_size = config["EBSOptions"]["Options"]["VolumeSize"]
        aei = config["AdvancedOptions"]["Options"]["rest.action.multi.allow_explicit_index"]
        ear = config["EncryptionAtRestOptions"]["Options"]["Enabled"]
        ne = config["NodeToNodeEncryptionOptions"]["Options"]["Enabled"]
        https = config["DomainEndpointOptions"]["Options"]["EnforceHTTPS"]
        fgac = config["AdvancedSecurityOptions"]["Options"]["Enabled"]
        internal_user_database = config["AdvancedSecurityOptions"]["Options"]["InternalUserDatabaseEnabled"]
        vpc = str(config["VPCOptions"]["Options"])
        access_policy = str(config["AccessPolicies"]["Options"])
        log_publish = str(config["LogPublishingOptions"]["Options"])
        warm_enabled=False
        if "WarmEnabled" in config["ElasticsearchClusterConfig"]["Options"]:
            warm_enabled = config["ElasticsearchClusterConfig"]["Options"]["WarmEnabled"]
        warm_instance_type = None
        if "WarmType" in config["ElasticsearchClusterConfig"]["Options"]:
            warm_instance_type = config["ElasticsearchClusterConfig"]["Options"]["WarmType"]
        warm_instance_count = None
        if "WarmCount" in config["ElasticsearchClusterConfig"]["Options"]:
            warm_instance_count = config["ElasticsearchClusterConfig"]["Options"]["WarmCount"]
        code_enabled = config["ElasticsearchClusterConfig"]["Options"]["ColdStorageOptions"]["Enabled"]
        auto_tune_enable = config["AutoTuneOptions"]["Status"]["State"]
        domain_config = DomainConfig(version, dn_intance_type, dn_instance_count,
                                     dedicated_master, dedicated_master_type, dedicated_master_count,
                                     dn_zone_awareness,dn_zone_number, ebs, ebs_volume_size,
                                     aei, ear, ne, https,fgac,internal_user_database,
                                     vpc,access_policy,log_publish,warm_enabled,warm_instance_type,warm_instance_count,
                                     code_enabled,auto_tune_enable)
        reccomendations = get_recommendation("domain_config.txt",domain_config)
        return reccomendations, domain_config

def gen_rec_from_rule(report_file,metadata_file, data):
    rec1 = get_recommendation_from_data(metadata_file, data)
    return rec1


def gen_cluster_settings(data_folder,report_file):
    with open(data_folder + "/cluster_settings_default.json") as f:
        data = json.load(f)
        recommendations = gen_rec_from_rule(report_file, "cluster_settings_default.json", data)
        return recommendations, data


def gen_cluster_health(data_folder,report_file):
    with open(data_folder + "/cluster_health.json") as f:
        data = json.load(f)
        recommendations = gen_rec_from_rule(report_file,"cluster_health.json",data)
        return recommendations, data

def gen_cluster_state(data_folder,report_file):
    with open(data_folder + "/cluster_state.json") as f:
        data = json.load(f)
        recommendations = gen_rec_from_rule(report_file,"cluster_state.json",data)
        cluster_templates = data["metadata"]["templates"]
        count = 0
        for template in cluster_templates:
            count += 1
        if count == 0:
            rec_entry = {"recommendation": "It is suggested to use template to create index"}
            rec_entry["rule_name"] = "dynamic-rule-cs-0"
            rec_entry["priority"] = 1
            rec_entry["observation"] = "It is better to use template to manage index"
            rec_entry["impact"] = "Operation"
            rec_entry["report_section"] = "cluster"
            recommendations["dynamic-rule-cs-0"] = rec_entry
        return recommendations, data

def gen_cluster_stats(data_folder,report_file):
    with open(data_folder + "/cluster_stats.json") as f:
        data = json.load(f)
        recommendations = gen_rec_from_rule(report_file, "cluster_stats.json", data)
        return recommendations, data

def gen_thread_pool(data_folder,report_file):
    with open(data_folder + "/thread_pool_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        return gen_rec_from_rule(report_file, "thread_pool_table.csv", data)

def gen_cluster_nodes(data_folder,report_file):
    recommendations={}
    #with open(data_folder + "/master.json") as f:
    #    data = json.load(f)
    #    gen_rec_from_rule(report_file, "master.json", data)

    #with open(data_folder + "/nodes.json") as f:
    #    data = json.load(f)
    #    gen_rec_from_rule(report_file, "nodes.json", data)

    with open(data_folder + "/nodes_cat_table.csv") as f:
        nodes_data = pd.read_csv(f, delim_whitespace=True)
        nodes_recs = gen_rec_from_rule(report_file, "nodes_cat_table.csv", nodes_data)
        if nodes_recs is not None:
            recommendations.update(nodes_recs)

    with open(data_folder + "/nodes_stats.json") as f:
        data = json.load(f)
        nodes_stats_recs = gen_rec_from_rule(report_file, "nodes_stats.json", data)
        if nodes_stats_recs is not None:
            recommendations.update(nodes_stats_recs)

    with open(data_folder + "/nodes_usage.json") as f:
        data = json.load(f)
        nodes_usage_recs = gen_rec_from_rule(report_file, "nodes_usage.json", data)
        if nodes_usage_recs is not None:
            recommendations.update(nodes_usage_recs)

    return recommendations, nodes_data


def gen_index(data_folder,report_file):
    with open(data_folder + "/indices_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        data = data[~(data["index"].str.startswith(".", na=False))]
        recommendations = gen_rec_from_rule(report_file, "indices_table.csv", data)
    return recommendations,data

def get_hot_index_data(data_folder,report_file):
    if os.path.exists(data_folder + "/hot_indices_table.csv"):
        with open(data_folder + "/hot_indices_table.csv") as f:
            data = pd.read_csv(f, delim_whitespace=True)
            data = data[~(data["index"].str.startswith(".", na=False))]
            return data

def get_warm_index_data(data_folder,report_file):
    if os.path.exists(data_folder + "/warm_indices_table.csv"):
        with open(data_folder + "/warm_indices_table.csv") as f:
            data = pd.read_csv(f, delim_whitespace=True)
            data = data[~(data["index"].str.startswith(".", na=False))]
            return data

def gen_field(data_folder,report_file):
    with open(data_folder + "/fielddata.json") as f:
        data = json.load(f)
        return gen_rec_from_rule(report_file, "fielddata.json", data)


def gen_mapping(data_folder,report_file):
    with open(data_folder + "/mapping.json") as f:
        data = json.load(f)
        return gen_rec_from_rule(report_file, "mapping.json", data)

def gen_allocation(data_folder,report_file):
    with open(data_folder + "/allocation_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        return gen_rec_from_rule(report_file, "allocation_table.csv", data)

def gen_shard(data_folder,report_file):
    with open(data_folder + "/shards_detail_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        data = data[~(data["index"].str.startswith(".", na=False))]
        unassigned_shards = data[data["state"] == "UNASSIGNED"]
        if len(unassigned_shards)>0:
            data=data[data["state"] != "UNASSIGNED"]
        data["store"] = data["store"].astype(int)
        recommendations = gen_rec_from_rule(report_file, "shards_detail_table.csv", data)

        if len(unassigned_shards)>0:
            data=data[data["state"] != "UNASSIGNED"]
            rec_entry = {"recommendation": "Primary or replica shards are skewed. It's better to balance the shards"}
            rec_entry["rule_name"] = "dynamic-rule-shard-2"
            rec_entry["priority"] = 1
            rec_entry[
                "observation"] = "Unassigned shards are found in the cluster. Unassigned shards(some of them):" + str(
                unassigned_shards["index"].tolist()[:20])
            rec_entry["impact"] = "Stability"
            rec_entry["report_section"] = "shard"
            recommendations["dynamic-rule-shard-2"] = rec_entry

        hot_shards_data = get_hot_shards_data(data_folder,report_file)
        if hot_shards_data is None or len(hot_shards_data) <=0:
            hot_shards_data = data
        if hot_shards_data is not None and len(hot_shards_data) > 0:
            grouped_data = hot_shards_data.groupby(by=["index", "prirep", "node"], as_index=False)["state"].agg({"total_shards": "count"})
            std_data = grouped_data.groupby(by=["index", "prirep"], as_index=False)["total_shards"].std().rename(columns={"total_shards":"shards_std"})
            skewed_shards = std_data[(std_data["shards_std"].notnull()) & (std_data["shards_std"] != 0.0)]
            if len(skewed_shards) > 0:
                rec_entry = {"recommendation": "Primary or replica shards are skewed. It's better to balance the shards"}
                rec_entry["rule_name"] = "dynamic-rule-shard-0"
                rec_entry["priority"] = 1
                rec_entry["observation"] = "It's better to balance shards among nodes for one index. Skewed index(some of them):" + str(skewed_shards["index"].tolist()[:20])
                # print("Skewed index:" + str(skewed_shards["index"].tolist()))
                # print(skewed_shards[["index","shards_std"]])
                rec_entry["impact"] = "Performance"
                rec_entry["report_section"] = "shard"
                recommendations["dynamic-rule-shard-0"] = rec_entry

        domain_recommendations, domain_config = gen_domain_config(data_folder, report_file)
        if domain_config.warm_enabled:
            warm_shards_data = get_warm_shards_data(data_folder, report_file)
            if len(warm_shards_data) > 0:
                warm_grouped_data = warm_shards_data.groupby(by=["index", "prirep", "node"], as_index=False)["state"].agg(
                    {"total_shards": "count"})
                warm_std_data = warm_grouped_data.groupby(by=["index", "prirep"], as_index=False)["total_shards"].std().rename(
                    columns={"total_shards": "shards_std"})
                warm_skewed_shards = warm_std_data[(warm_std_data["shards_std"].notnull()) & (warm_std_data["shards_std"] != 0.0)]
                if len(warm_skewed_shards) > 0:
                    rec_entry = {"recommendation": "Primary or replica shards are skewed in UltraWarm nodes. It's better to balance the shards"}
                    rec_entry["rule_name"] = "dynamic-rule-shard-1"
                    rec_entry["priority"] = 1
                    rec_entry[
                        "observation"] = "It's better to balance shards among UltraWarm nodes for one index. Skewed index(some of them):" + str(
                        warm_skewed_shards["index"].tolist()[:20])
                    # print("Skewed index:" + str(skewed_shards["index"].tolist()))
                    # print(skewed_shards[["index","shards_std"]])
                    rec_entry["impact"] = "Performance"
                    rec_entry["report_section"] = "shard"
                    recommendations["dynamic-rule-shard-1"] = rec_entry
        return recommendations,data

def get_hot_shards_data(data_folder,report_file):
    if os.path.exists(data_folder + "/hot_shards_table.csv"):
        with open(data_folder + "/hot_shards_table.csv") as f:
            data = pd.read_csv(f, delim_whitespace=True)
            data = data[~(data["index"].str.startswith(".", na=False))]
            return data

def get_warm_shards_data(data_folder,report_file):
    if os.path.exists(data_folder + "/warm_shards_table.csv"):
        with open(data_folder + "/warm_shards_table.csv") as f:
            data = pd.read_csv(f, delim_whitespace=True)
            data = data[~(data["index"].str.startswith(".", na=False))]
            return data

def gen_cross_section(data_folder, recommendations,report_file):
    domain_recommendations, domain_config = gen_domain_config(data_folder, report_file)
    nodes_recs, nodes_data = gen_cluster_nodes(data_folder, report_file)
    instance_count = domain_config.instance_count

    hot_index_data = get_hot_index_data(data_folder,report_file)
    if hot_index_data is None or len(hot_index_data) <= 0:
        recs, hot_index_data = gen_index(data_folder,report_file)
    hot_skew_index = hot_index_data[hot_index_data["pri"]%instance_count !=0 ]["index"]
    if len(hot_skew_index) > 0:
        rec_entry = {"recommendation": "It is suggested to balance primary shards evenly among nodes. Skewed index(some of them):"
                                       + str(hot_skew_index.tolist()[:20])}
        rec_entry["rule_name"] = "dynamic-rule-index-0"
        rec_entry["priority"] = 1
        rec_entry["observation"] = "It is better to balance primary shards evenly among nodes"
        rec_entry["impact"] = "Performance"
        rec_entry["report_section"] = "index"
        recommendations["dynamic-rule-index-0"] = rec_entry
    if domain_config.warm_enabled:
        warm_instance_count = domain_config.warm_instance_count
        warm_index_data = get_warm_index_data(data_folder,report_file)
        if warm_index_data is not None and len(warm_index_data) > 0:
            warm_skew_index = warm_index_data[warm_index_data["pri"] % warm_instance_count != 0]["index"]
            if len(warm_skew_index) > 0:
                rec_entry = {"recommendation": "It is suggested to balance primary shards evenly among UltraWarm nodes. Skewed index(some of them):"
                                               + str(warm_skew_index.tolist()[:20])}
                rec_entry["rule_name"] = "dynamic-rule-index-1"
                rec_entry["priority"] = 1
                rec_entry["observation"] = "It is better to balance primary shards evenly among UltraWarm nodes"
                rec_entry["impact"] = "Performance"
                rec_entry["report_section"] = "index"
                recommendations["dynamic-rule-index-1"] = rec_entry

    node_instance_type = domain_config.instance_type
    vcpu = get_vcpu(node_instance_type)
    busy_cpu = nodes_data[nodes_data["load_15m"] > vcpu]
    my_busy_cpu = busy_cpu[["name","load_15m"]]
    if len(busy_cpu) > 0:
        rec_entry = {"recommendation": "CPU load in the past 15 minutes is high. Nodes(some of them): " + str(busy_cpu["name"].tolist()[:20])}
        rec_entry["rule_name"] = "dynamic-rule-node-0"
        rec_entry["priority"] = 1
        rec_entry["observation"] = "CPU load in the past 15 minutes is high"
        rec_entry["impact"] = "Performance"
        rec_entry["report_section"] = "node"
        recommendations["dynamic-rule-node-0"] = rec_entry


def get_vcpu(instance_type):
    instance_type_parts = instance_type.split(".")
    vcpu=2
    if len(instance_type_parts) == 3:
        pos = instance_type_parts[1].find("xlarge")
        if pos > 0:
            scale = int(instance_type_parts[1][0:pos])
            vcpu = scale * 4
        else:
            pos2 = instance_type_parts[1].find("large")
            if pos2 == 0:
                vcpu = 4
    return vcpu



def generate_recommendation(data_folder, report_file):
    recommendations = {}
    # Cluster level
    domain_recommendations, domain_config = gen_domain_config(data_folder, report_file)
    if domain_recommendations is not None:
        recommendations.update(domain_recommendations)
    cluster_settings_recommendations,cluster_settings = gen_cluster_settings(data_folder, report_file)
    if cluster_settings_recommendations is not None:
        recommendations.update(cluster_settings_recommendations)
    cluster_health_recommendations,cluster_health = gen_cluster_health(data_folder, report_file)
    if cluster_health_recommendations is not None:
        recommendations.update(cluster_health_recommendations)
    cluster_state_recommendations,cluster_state = gen_cluster_state(data_folder, report_file)
    if cluster_state_recommendations is not None:
        recommendations.update(cluster_state_recommendations)
    cluster_stats_recommendations, cluster_stats = gen_cluster_stats(data_folder, report_file)
    if cluster_stats_recommendations is not None:
        recommendations.update(cluster_stats_recommendations)
    # Nodes level
    nodes_recommendations, nodes_data = gen_cluster_nodes(data_folder, report_file)
    if nodes_recommendations is not None:
        recommendations.update(nodes_recommendations)
    # Index level
    index_recommendations,index_data = gen_index(data_folder, report_file)
    if index_recommendations is not None:
        recommendations.update(index_recommendations)
    # Shard level
    shard_recommendations,shard_data = gen_shard(data_folder, report_file)
    if shard_recommendations is not None:
        recommendations.update(shard_recommendations)
    allocation_recommendations = gen_allocation(data_folder, report_file)
    if allocation_recommendations is not None:
        recommendations.update(allocation_recommendations)
    # Field and mapping
    # field_recommendations = gen_field(data_folder, report_file)
    # if field_recommendations is not None:
    #   recommendations.update(field_recommendations)
    # mapping_recommendations = gen_mapping(data_folder, report_file)
    #if mapping_recommendations is not None:
    #   recommendations.update(mapping_recommendations)
    # Thread pool
    thread_pool_recommendations = gen_thread_pool(data_folder, report_file)
    if thread_pool_recommendations is not None:
        recommendations.update(thread_pool_recommendations)
    gen_cross_section(data_folder,recommendations,report_file)
    return recommendations

